// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_loopback.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Jacques Briat, Thierry Gautier, Cyrille Martin
// ==========================================================================
#ifndef _ACOM_LOOPBACK_H_
#define _ACOM_LOOPBACK_H_

#include <inuk/acom_network.h>


// ACOM namespace
namespace ACOM_NET_NAMESPACE {


// ---------------------------------------------------------------------------
// LoopBack class
// ---------------------------------------------------------------------------
/**@name LoopBack network
The network local to a node. 

*/
class LoopBack : public Network
{
protected :

  /** size of cached free blocks
  */
  static const int _block_threshold=2; // size of cached free blocks
  /** size of cached free blocks
  */
  static const int _iovect_threshold=2; // size of cached free blocks
  /** size of cached free blocks
  */
  static const int _request_threshold=2; // size of cached free blocks
  /** iovect entries in a cell
  */
  static const int _nb_entries=10; // iovect entries in a cell 
  /** packed block size
  */
  static const int _block_sz = 100; // packed block size

  /** Logical clock for timestamping request
  */
  PTH::AtomicCounter _comClock;

  
  /** byte buffer  allocator for sending 
  */
  ByteCellAllocator _bA;
  /** iovec allocator for sending 
  */
  IOVectCellAllocator _iovA;
  /** byte buffer  allocator for receive
  */
  ByteCellAllocator _bA_RCV;
  /** iovec allocator  for receive
  */
  IOVectCellAllocator _iovA_RCV;

  /** init all the allocation stuff 
  */
  void initAllocators(){ 
    _comClock=0;
    _reqA.init(_request_threshold);  
    _bA.init(_block_sz, _block_threshold  );
    _bA_RCV.init(_block_sz, _block_threshold  );
    _iovA.init(
      sizeof(IOVectCell)+(_nb_entries*sizeof(IOVectEntry)) , 
      _iovect_threshold    );
    _iovA_RCV.init(
      sizeof(IOVectCell)+(_nb_entries*sizeof(IOVectEntry)) , 
      _iovect_threshold    );
  };
  /**  The Loop request 
  */
  class LoopRequest : public Network::Request, public PTH::Linkable<LoopRequest> {
  protected :
    bool init(size_t bsz, ByteCellAllocator* bufa,
              int iovsz, IOVectCellAllocator * iova,
              Network * manager)
    { return Network::Request::init(bsz,bufa,iovsz, iova,manager);}  
    friend class LoopBack;
  };
  /** simulate a post on a loopback link
  */
  void simulatePost( LoopRequest* req );
  void simulateReceive(LoopRequest* req , size_t* sz);

  /** request allocator 
  */

  typedef PTH::SyncCellAllocator<LoopRequest> RequestAllocator ;
 RequestAllocator _reqA;

  /**  The Loop receipt 
  */
  class LoopReceipt : public Network::Receipt {
  protected :
    bool init( ByteSeq * bs , IOVectEntry * iov, size_t sz, Network* manager)
    { return Network::Receipt::init(bs,iov , sz,manager);}
    bool init( IOVectSeq * bs , IOVectEntry * iov, size_t sz, Network* manager)
    { return Network::Receipt::init(bs,iov , sz,manager);}
    friend class LoopBack;
  };

  /** the only channel class
  */
  class LoopChannel : public Network::OutChannel {
  public :
    void post( Network::Request* req )throw (ComFailure)  
    { get_net()->post(req,this);};
  protected  :
    void init(Network*  net, LocalNodeNo peer)
    { OutChannel::init(net,peer);  }
  friend class LoopBack;
  };
  /**  only one channel is needed
  */
  LoopChannel _ch; // the only channel 
  /** The only peer is this node !!
  */
  NodeDesc _peer_loop;

public:
  LoopBack():Network(),   _comClock(0),   _bA(), _iovA(), _reqA()  {}
    

  virtual ~LoopBack() {}

  // ---------------------------------------------------------------------------
  // Network interface
  // ---------------------------------------------------------------------------
public:
  
  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help. 
  */
  int initialize( int* argc, char** argv[] )throw (NetworkError){
    static bool once=false; if (once) return 0; once=true;
    initAllocators();
    PTH::Pth::initialize();
    set_name("LOOPBACK_NET");
    _node_label= "THIS_NODE";
    _node_rank=0; _node_count=1;
    _ch.init(this,_node_rank);
    _peer_loop.init(_node_rank,_node_label ); 
    _known_nodes[_node_rank]=&_peer_loop;
    _default_routes[_node_rank]=&_ch;
    return 0;
  };

  int initialize( int* argc, char** argv[], acom::Network* netInit )
    throw (NetworkError) 
  { return initialize(argc,argv); }

  /** Second stage of the initialisation process.
      End of the first stage of initialisation : all local initialisations
      are done, communication could occurs after the second stage (after
      the call to initcommit) or during the call of initcommit in an other
      module. 
  */
  int commit ()throw (NetworkError) {PTH::Pth::commit();return 0;};

  /** Terminate the net... 
  */
  int terminate ()throw (NetworkError) {PTH::Pth::terminate();return 0;};


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Allocate a new request.
  */
   Request* allocate_request () throw (BadAlloc) { LoopRequest* req ;
  
     req = _reqA.alloc() ; 
      if(!req) throw BadAlloc("[[LoopBack::allocate_request ::]] fail")  ;
      if(!req->init( _block_sz,& _bA,_nb_entries ,& _iovA,(Network*)this ))
        {_reqA.free(req);throw BadAlloc("[[LoopBack::allocate_request ::]] fail")  ;}
      return req;
  }

  /** free request descriptor when unused
  */
   void deallocate_request( Request* rq ) throw (InvalidArgument) {
    LoopRequest* req = (LoopRequest*)rq;
    if(req){
      if(req->get_net()!=this) 
        throw InvalidArgument("[[LoopBack::deallocate_request ::]] not a loopback request")  ;
      else{
        _reqA.free(req); 
        }
      }
    else throw InvalidArgument("[[LoopBack::deallocate_request ::]] null request")  ;
  };

  /** Cancel a request allocated by allocate_request ( true == cancel done)
      Cancelation of request depend on state of progress
      @param req (IN)
  */
  bool cancel_request( Request* req ) 
    throw (InvalidArgument) 
  { return false; }


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Allocate  an immediate buffer of sz bytes in network memory
      @param  sz (IN) in bytes 
  */
   void* allocate_data( size_t sz ) 
       throw (BadAlloc) { return ::malloc(sz);};

  /** Deallocate a previously allocated immediate buffer  in network memory
      @param  data (IN) byte block to deallocate
  */
   void deallocate_data(  void* data ) 
       throw (InvalidArgument) { ::free(data);};

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Add to  network memory a part of user memory
      @param  begin (IN)IN) beginning address of the memory block 
      @param  sz sz   (IN)in bytes of the memory block 
      \warning The zone is mapped. The really mapped zone is the rounding of 
      the zone in system page size unit. At this level, no registering of 
      concurrent mappings or overlapping of zones are done.
  */
  void  map_zone( void * begin, size_t sz) throw (BadAlloc) {}

  /** Retract from network memory a part of user memory 
      @param  begin (IN) beginning address of the memory block 
      @param  sz   (IN) in bytes of the memory block 
      \warning The zone is unmapped. At this level, no control of overlapping 
      of mapping or multiples mapping of the same zone are done..   
  */
   void  umap_zone( void * begin, size_t sz)throw (InvalidArgument) {}


  /** plug the current node to a  network address
       @param   nodelabel(IN) network label of remote node
      \warning ajouter les aliasing d'adresse de noeud struct  ( A voir  partir IP/port, url corba ..) 
  */
   void bind_node( const std::string& nodelabel   ) 
       throw(NetworkError) {    };

  /** set  the identifier of the current node in the Network
  */
  void set_rank(LocalNodeNo n){ }

     
  /** add a new node to the network
    @param  nodename(IN) network identification of remote node
    @return local node number of the new node
  \warning MUST fail when multiple add of same addreses

   \warning ajouter les aliasing d'adresse de noeud   ( A voir  partir fd, IP/port, url corba ..)
  */
   void  add_node( const std::string& nodelabel, LocalNodeNo  peer ) 
       throw(NetworkError) { throw NetworkError("[[LoopBack::add_node ::]]  null op");} ;

   
  /** del a previously added node to the network
      @param  peer  local node Number of remote node to disconnect
      \warning ajouter les aliasing d'adresse de noeud     ( A voir  partir IP/port, url corba ..)
  */
  void  del_node( LocalNodeNo  peer  ) 
      throw(NetworkError) { throw NetworkError("[[LoopBack::add_node ::]]  null op");} ;



  // ---------------------------------------------------------------------------    
  // ---------------------------------------------------------------------------    
  /** Return the route (channel) to send data to LocalNodeNo peer
      @param   peer(IN) remote node id
  */
   OutChannel* get_default_route( LocalNodeNo peer ) 
       throw(NetworkError) { return &_ch;};

/** Set a channel as a default channel to a destination node
      @param   ch default channel 
  */
   void set_default_route(  OutChannel* ch  ) 
       throw(NetworkError) {};


  /** Open a channel between local node  to remote peer
      @param   peer(IN) remote node id
      \warning ajouter les aliasing d'adresse de noeud  
  */
   OutChannel* open_channel( LocalNodeNo peer ) 
       throw(ComFailure){return &_ch;};

  /** Close a channel between local node  to remote peer
      @param   ch (IN)channel to close 
      \warning ajouter les aliasing d'adresse de noeud  
  */
   void close_channel( OutChannel* ch ) 
       throw(ComFailure) {};


  // ---------------------------------------------------------------------------  
  // ---------------------------------------------------------------------------  

/** broadcast to nodes (AM,AWS,WS protocols) 
      @param   req  request (IN) 
      @param   ch  channel  (IN) 
  */
  void broadcast( Network::Request* req, OutChannel* ch=0 ) 
        throw (ComFailure){
    if(req){
      if(req->get_net()!=this) 
        throw ComFailure(ComFailure::BAD_REQUEST)  ;
      else
        simulatePost((LoopRequest*)req); return;
      }
    else throw ComFailure(ComFailure::BAD_REQUEST)  ;
  };

  /** neighbour_cast to nodes (AM,AWS,WS protocols) 
      @param   req  request (IN) 
      @param   ch  channel  (IN) 
  */
  
  void neighbour_cast( Network::Request* req ,OutChannel* ch=0) 
       throw (ComFailure) { 
     if(req){
      if(req->get_net()!=this) 
        throw ComFailure(ComFailure::BAD_REQUEST)  ;
      else
        simulatePost((LoopRequest*)req); return;
      }
    else throw ComFailure(ComFailure::BAD_REQUEST)  ;
  };
/** send  to a node (AM,AWS,WS protocols) 
      @param   req  request (IN) 
       @param   ch  channel  (IN) 
  */
  void  post( Network::Request* req ,OutChannel* ch=0) 
       throw (ComFailure) { 
   if(req){
      if(req->get_net()!=this) 
        throw ComFailure(ComFailure::BAD_REQUEST)  ;
      else
        simulatePost((LoopRequest*)req); return;
      }
    else throw ComFailure(ComFailure::BAD_REQUEST)  ;
  };
  // ---------------------------------------------------------------------------  
  // ---------------------------------------------------------------------------  
  /** Network specific barrier
      \warning  parameters count, time out etc ...( voir avec Cyrille)
  */ 
   void barrier()
       throw (ComFailure) { };

  



}; // -- end class Network


} // end namespace ATH...
#endif 
