// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_network.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Jacques Briat, Thierry Gautier, Cyrille Martin
// ==========================================================================
#ifndef _ACOM_NETWORK_H_
#define _ACOM_NETWORK_H_

#include <inuk/acom_netbase.h>
#include <inuk/acom_netclient.h>
#include <inuk/acom_netserver.h>
#include <inuk/acom_netadmin.h>

//@Include: acom_netconfig.h acom_comerror.h acom_types.h

// ACOM namespace
namespace ACOM_NET_NAMESPACE {


// ---------------------------------------------------------------------------
// Network class
// ---------------------------------------------------------------------------
/**@name Network
The virtual API for new Inuktitut interface Network.
An object of type Network defines types, data for the connection of the 
process to a particular Network. Communications are based on requests : 
a Network object can allocate request object that could be post into the 
net in different direction (IN == recv message or OUT == send message), 
using different mode (copy data before sending, free data at completion,...). 
Once the request is send/recv the notification 'notify' method is called.
*/
class Network : public NetworkBase, 
                public NetworkClient, 
                public NetworkServer, 
                public NetworkAdmin {
public:
  typedef NetworkBase::LocalNodeNo  LocalNodeNo;
  typedef NetworkBase::BaseRequest  BaseRequest;
  typedef NetworkClient::Request    Request;
  typedef NetworkClient::OutChannel OutChannel;
  typedef NetworkServer::Receipt    Receipt;

  /** Default Cstor
  */
  Network();
  
  /** Destructor
  */
  virtual ~Network();
  
  /** broadcast to nodes (AM,AWS,WS protocols) 
      @param   req  request (IN) 
      @param   ch  channel  (IN) 
  */
  virtual void broadcast( Request* req , OutChannel* ch=0) 
       throw (ComFailure);

  /** neighbour_cast to nodes (AM,AWS,WS protocols) 
      @param   req  request (IN) 
      @param   ch  channel  (IN) 
  */
  virtual void neighbour_cast( Request* req, OutChannel* ch=0 ) 
       throw (ComFailure);

  /** Network specific barrier
      \warning  parameters count, time out etc ...( voir avec Cyrille)
  */ 
  virtual void barrier()
       throw (ComFailure);

}; // -- end class Network


/*
 * inline definition
 *
 */
/* should be put because it need definition of Network */

#if 0 /* DEPRECATED */
inline void* NetworkBase::BaseRequest::allocate_extended_header( size_t s)
  throw (BadAlloc)
{
  _size_eh = s;
  return _ext_header = _net->allocate_data( s );
}
#endif

inline void NetworkServer::CallContext::release()
{
  ACOM_ASSERT_DEBUG(network !=0, "bad assertion");
  network->release_callcontext( this );
}


} // end namespace ATH...
#endif 
