// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_protocol.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Jacques Briat, Thierry Gautier, Cyrille Martin
// ==========================================================================
#ifndef _ACOM_PROTOCOL_H_
#define _ACOM_PROTOCOL_H_

#ifndef WIN32
#include <inuk/acom_netconfig.h>
#endif
#include <inuk/acom_iovect.h>
#include <inuk/acom_types.h>
#include <inuk/acom_network.h>
#include <inuk/acom_channel.h>

//@Include: acom_netconfig.h acom_network.h 

// ACOM namespace
namespace ACOM_NET_NAMESPACE {


// ---------------------------------------------------------------------------
// Implementation of protocol with service over Network 
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
/**@name Message
Three kind of Message exist. The ActiveMessage and WriteAndSignal and 
AllocateWriteAndSignal protocols. Message is base class for these kinds
of message.
@see ActiveMessage, WriteAndSignal and AllocateWriteAndSignal
*/
class Message {
public:
  /** Constructor of a message
  */
  Message() : _req(0)
  { }

  /** Constructor of a message, protocol is undefined
  */
  Message( Network* net, Network::ProtocolNo p, Service* serv, Callback* cbk, int usecount=1) 
  {
    _req = net->allocate_request( );  
    _req->_used_ct=usecount;
    _req->_f_cbk=cbk;
    _req->_header.id= serv->get_id();
    _req->_header.protocol= p;
  }
  
  /** Destructor with given size for the header:
  */
  ~Message()
  { _req=0; }

  /**  Make transparent access to Request
  */
  operator Network::Request*() 
  { return _req; }

  /**  Make transparent access to Request
  */
  operator const Network::Request*() const
  { return _req; }

  /**  Make transparent access to Request
  */
  Network::Request* operator->() const 
  { return _req; }
  
  /** Return access to request object
  */
  Network::Request* get_request() const 
  { return _req; }

  /** (re)direct the message to a (new)  channel
  */
  void post( Network::OutChannel* ch) {
    ch->post(_req);
  }

#if 0 /* DEPRECATED */
  /** (re)direct the message to a (new)  channel
  */
  void post( Network::OutChannel& ch) {
    //std::cout << "REQ POST CH&"<<_req << "\n";
    ch.post(_req);
    //std::cout << "END REQ POST "<< "\n";
  }

  /** (re)direct the message to a (new)  channel
  */
  void operator()( Network::OutChannel* ch ) 
  {
    //std::cout << "REQ POST CH () "<<_req << "\n";
    ch->post(_req); 
    //std::cout << "END REQ POST "<< "\n";
  }
#endif

protected:
  Network::Request* _req; 
};


// ---------------------------------------------------------------------------
/**@name AMMessage
Three kind of ProtocolService exist. The ActiveMessage and WriteAndSignal
and AllocateWriteAndSignal protocols. Given a service object 'service' of type SERVICE needing
an immediate argument "value_arg" and  a variable array argument "val_array" a call to 'service' should be:
\code
AMMessage am (net, service, cbk);
am->pack(&user_tag,sizeof(user_tag));
am->pack(&val_array_size, sizeof(int));
am->put(&valarray, val_array_size);
ch.post(am);
\endcode
@see Service
*/
class AMMessage : public Message {
public:
  /** Constructor with given size for the header:
  */
  AMMessage( Network* net, Service* serv, Callback* cbk, int usecount=1) 
   : Message( net, Network::P_AM, serv, cbk, usecount)
  { }
}; // -- end class AMMessage


// ---------------------------------------------------------------------------
/**@name WSMessage
Three kind of ProtocolService exist. The ActiveMessage and WriteAndSignal
and AllocateWriteAndSignal protocols. Given a service object 'service' of type SERVICE needing
an immediate argument "value_arg" and  a variable array argument "val_array" a call to 'service' should be:
\code
WSMessage am (net, service, cbk);
am->pack(&user_tag,sizeof(user_tag));
am->pack(&val_array_size, sizeof(int));
am->put(&valarray, val_array_size);
ch.post(am);
\endcode
@see Service
*/
class WSMessage : public Message {
public:
  /** Constructor with given size for the header:
  */
  WSMessage( Network* net, Service* serv, Callback* cbk, int usecount=1) 
   : Message( net, Network::P_WS, serv, cbk, usecount)
  { }
}; // -- end class WSMessage


// ---------------------------------------------------------------------------
/**@name AWSMessage
Three kind of ProtocolService exist. The ActiveMessage and WriteAndSignal
and AllocateWriteAndSignal protocols. Given a service object 'service' of type SERVICE needing
an immediate argument "value_arg" and  a variable array argument "val_array" a call to 'service' should be:
\code
AWSMessage am (net, service, cbk);
am->pack(&user_tag,sizeof(user_tag));
am->pack(&val_array_size, sizeof(int));
am->put(&valarray, val_array_size);
ch.post(am);
\endcode
@see Service
*/
class AWSMessage : public Message {
public:
  /** Constructor with given size for the header:
  */
  AWSMessage ( Network* net, Service* serv, Callback* cbk, int usecount=1) 
   : Message( net, Network::P_AWS, serv, cbk, usecount)
  { }
}; // -- end class AWSMessage



// ---------------------------------------------------------------------------
// Implementation of ActiveMessage protocol
// ---------------------------------------------------------------------------
/**@name ActiveMessage
*/
template<class SERV>
class ActiveMessage : public AMMessage {
public:
  ActiveMessage (  Network* net, SERV* serv, Callback* cbk, int user_count = 1) 
   : AMMessage(net, serv, cbk, user_count )
  {}
};


// ---------------------------------------------------------------------------
// Implementation of WriteAndSignal protocol
// ---------------------------------------------------------------------------
/**@name WriteAndSignal
*/
template<class SERV>
class WriteAndSignal : public WSMessage {
public:
  WriteAndSignal (  Network* net, SERV* serv, Callback* cbk, int user_count = 1) 
   : WSMessage(net, serv, cbk, user_count )
  {}
};


// ---------------------------------------------------------------------------
// Implementation of WriteAndSignal protocol
// ---------------------------------------------------------------------------
/**@name AllocWriteAndSignal
  Does not need get_iovect : it was allocated by the network !
*/
template<class SERV>
class AllocWriteAndSignal : public AWSMessage {
public:
  AllocWriteAndSignal
    (  Network* net, SERV* serv, Callback* cbk, int user_count = 1 )
   : AWSMessage(net, serv, cbk, user_count) {}
};

} // end namespace ACOM...
#endif 
