// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_stack_pool.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// - Utility for manage pool of stacks for service call.
// Should be inclued with location of inukthread.
// ==========================================================================
#ifndef _ACOM_POOL_STACK_H_ 
#define _ACOM_POOL_STACK_H_

#include <inuk/acom_netconfig.h>
#include <inuk/acom_types.h>
#include <stddef.h>      // - for size_t

//@Include: acom_netconfig.h  

// A0 namespace
namespace ACOM_NET_NAMESPACE {


class Stack {
public:
  /** Stack
  */
  Stack( );

  /** Stack
  */
  void init( size_t capacity );

  /** Destructor Stack
  */
  ~Stack( );

  /// destroy
  void destroy();

  /// return the count of free stack 
  int count_stacks() const;

  /// acquire a new stack, m is the mutex that is taken to ensure 
  /// mutal exclusion
  void acquire( void*&, size_t& sz, AT::Mutex& m_release );

  /// acquire a new stack
  void release( void*& );

protected:
  int                _stack_limit;
  size_t             _stack_size;
  std::vector<void*> _stacks;
  volatile int       _sig;
  std::list<void*>   _free_stacks;
  AT::Condition      _cond;
};




/*
 * inline definitions
 */
inline StackPool::StackPool()
{
  _stacks.clear();
  _free_stacks.clear();
}

  /// StackPool
inline void StackPool::init
( size_t stack_size, int stack_capacity, int max_limit )
{
  _stack_size = stack_size;
  _stack_limit = max_limit;
  _stacks.resize( stack_capacity );
  _free_stacks.clear();
  for (int i=0; i<stack_capacity; ++i)
  {
    _stacks[i] = new char[_stack_size];
    _free_stacks.push_back(_stacks[i]);
  }
  _cond.init();
}

  /// ~StackPool
inline void StackPool::destroy()
{
  if (_free_stacks.size() != _stacks.size())
    throw Runtime("[StackPool::~StackPool] some stacks are not released");
  for (int i=0; i< _stacks.size(); ++i)
  {
    delete [] (char*)_stacks[i];
  } 
  _stacks.clear();
  _free_stacks.clear();
}

inline StackPool::~StackPool( )
{
  destroy();
}

  /// return the count of free stack 
inline int StackPool::count_stacks() const
{
  return _stacks.size();
}

  /// acquire a new stack
inline void StackPool::acquire( void*& st, size_t& sz, AT::Mutex& m_release )
{
  int stack_id =0;
  sz = _stack_size;
  while (_free_stacks.empty()) 
  {
    stack_id = _stacks.size();
    if ((_stack_limit !=-1) && (stack_id < _stack_limit)) {
      _stacks.resize( stack_id+1 );
      st = _stacks[stack_id] = new char[_stack_size]; 
      return;
    }
    else {
      ++_sig;
      _cond.wait( m_release );
    }
  }
  st = _free_stacks.front(); 
  _free_stacks.pop_front(); 
  return;
}

  /// acquire a new stack
inline void StackPool::release( void*& st )
{
  _free_stacks.push_front( st );
  st = 0;
  if (_sig >0) {
    --_sig;
    _cond.signal();
  }
}
 
} // namespace ACOM_NET_NAMESPACE 

#endif
