// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_log.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// $Id: pth_log.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ===================================================================

#ifndef _PTHlog_H_
#define _PTHlog_H_

/*!
\file  
\brief Log file management   
\author  J. Briat T. Gautier 
\date 2001

This file contains :
	- the log file class   PTH::Logger 
	- the Inuktitut Log management class PTH::InukLogger
	- General Log macros LOG() _LOG
*/





#include <inuk/pth_ns.h>
#include <iostream>
#include <string>
#include <fstream>
//#include <cstdlib>
#include <inuk/pth_macro.h>
#include <inuk/pth_error.h>
#include <inuk/pth_pthreadsync.h>

namespace PTH
{

/**
\brief A  Logger object class

A  logger  handles two log files. 

It is possible to log on the two files, to one of them or none. Enabling/disabling  a log file may be done dynamically. 

\warning  logging is thread unsafe if locking not explicitly done. see LOG() macros.

*/

class  Logger : public Mutex
{     
  /// the output channel pointer to file
  std::ostream  *_clog[2] ;
	bool _enabled[2];
public:
  /// Constructor
  Logger()throw ( PosixError )
   :  Mutex(),_enabled(){};
	/// initialize logger
	void init(){Mutex::init_notrace();};
  /// Destructor
  ~Logger() throw () {};
  /// map  one of the 2 log files   to a stream
  void bind(int i, std::ostream * fl)throw (){  _clog[i]=fl;};
  /// unmap   one of the 2 log files  to a stream
  void unbind(int i)throw (){_clog[i]=0; _enabled[i]=false; };
  /// output operator 
  template < class T > 
  Logger& operator<<( const T& yy)
  {
    if((_enabled[0])&& _clog[0]  )  * _clog[0] << yy;
    if((_enabled[1]) &&  _clog[1] )* _clog[1] << yy;
    return *this;
  }
	/// output operator 
  Logger& operator<<( std::ostream& (*fnc)(std::ostream&) )
  { 
    if((_enabled[0])&& _clog[0] ) * _clog[0] << fnc;
    if((_enabled[1]) &&  _clog[1]) * _clog[1] << fnc;
    return *this;
  }
  /// enable one of the 2 log files 
  void enable(int i){
    _enabled[i] |= true;
  }
  /// disable one of the 2 log files
  void disable(int i){_enabled[i] |= false;}
 /// flush the 2 log files
	void flush(){
		if(_clog[0] )  *_clog[0] << std::flush;;
    if( _clog[1])  *_clog[1]<< std::flush;
};

};

/**
\brief the log service management

  This is thedefault Inuktitut logger  initiated at init time (see PTH::System class)

  \warning If  To  log events in a log file in a thread-aware manner, it is necessary
   to commit the Pthread + Mutex  service before the Log service.

*/

extern  Logger* InukLogger;


} // end namespace


//----
// Log  macros  
//        LOG  thread aware log on log file
//
//        _LOG  thread unsafe log on log file
//               ( to be used only in internals and panic situation )
//
//
//  
//----

/**
\brief End of  log  recording

\warning Intermediate macro. For  internal use only.

*/ 
#define _ENDLOG_(Y) Y<< "\n"<< std::flush; }

/**
\brief Beginning of  thread-aware log.

This macro starts log recording. The lock associated with a PTH::InukLogger object is used for mutual exclusion. 
*/
#define LOG {CRITICAL_NOTRACE(PTH::InukLogger);*PTH::InukLogger << _ENDLOG_


/**
\brief Beginning of thread-unsafe log 

This macro starts log recording.
*/ 
#define _LOG {*PTH::InukLogger << _ENDLOG_

/// Inhibits log
#define NOLOG(X) 
/// Inhibits log
#define _NOLOG(X) 

#endif  // _PTHlog_H_
