 // ==========================================================================
// $Source: /home/cvs/Taktuk2/Psocket/src/sock_addr.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file for more information
// Author: J Briat T.Gautier
// Header file for Inuktitut/Thread/Pthread
//
//
//
// ==========================================================================
#ifndef __SOCK_SOCKADDR_H_
#define __SOCK_SOCKADDR_H_

#include <inuk/sock_dep.h>

#include <string>

// for sockaddr_in
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <string>
#include <inuk/sock_netaddr.h>

namespace SOCK {

class SocketClient;
class SocketServer;

class Address : public NetAddress {
public:
/*! Constructor, no initialization
 */
  Address();

  /** copy-constructor,initialize with specified address
   */
  Address(const Address& a);

  /** Destructor
   */
  ~Address();

  /** Assignement
   */
  Address& operator=(const Address& a);

  /**
   */
  int getFamily() const;

  /** initialize with host
   *  \return reference to initialized object
   */
  Address& setByName( const std::string& host );

  /** return the address of host given by name
   *  \return reference to initialized object
   */
  static Address getByName( const std::string& host );

  /**  Return the hostname of this address
   *  \return _host
   */
  const std::string& getHostName() const;

  /** get address object for localhost
   *  \return Address object for localhost
   */
  static const Address& getLocalHost();

  /** return the adress (ip+port)
   *  \return as string
   */
  std::string getURL() const;

  /** return description of last error
   *  \return error as string
   */
  std::string getLastError() const;

  /** return the last error no
   *  \return error number
   */
  int getLastErrorNo() const;

  /** return if object/ip is valid.
   *  \return true if host/IP is valid
   *  \warning if isValid() returns false you can't rely on the results of
   *           other get functions in Address
   */
  bool isValid() const;

  /** print address
   */
  std::ostream& print(std::ostream& ) const;

  /** set port of address
   *  \warning changing the port after connecting will not change the port your
   *           connection is on, but modify data wich will lead to inconsistency
   */
  void setPort(int p);

  /** return port number of address
   *  \return _port
   */
  int getPort() const;

protected: 
  /** return host address information
   *  \return sockaddr_in struct (system dependent structure to store address)
   *  \warning this is data inside the class, be carefull not to overwrite !
   */
  const sockaddr_in* getSockAddr_in() const;
  sockaddr_in* getSockAddr_in();
  const sockaddr* getSockAddr() const;
  sockaddr* getSockAddr();

  /**  
   *   \return the size of the addr
   */
  int getSockAddrSize() const;

private:
  bool          _valid;
  int           _lastError;
  std::string   _host;
  int           _port;
  sockaddr_in   _sockaddr;

  //unsigned char _addr[4];

  friend class Socket;
  friend class SocketClient;
  friend class SocketServer;
};

inline std::ostream& operator << (std::ostream& o, const Address& a )
{
  return a.print(o);
}

} // - namespace SOCK 

#endif

