// ==========================================================================
// $Source: /home/cvs/Taktuk2/Psocket/src/sock_basic.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file for more information
// Author: J Briat T.Gautier
// Header file for Inuktitut/Thread/Pthread
//
//
//
// ==========================================================================
#ifndef __SOCK_BASIC_H_
#define __SOCK_BASIC_H_

#include <inuk/sock_addr.h>


namespace SOCK {


/**
*/
class Socket {
private:
  /** Recopy Constructor
  */
  Socket(const Socket& s) {};

  /** Assignment operator
  */
  Socket& operator=(const Socket& s);

public:
  /** Default constructor
  */
  Socket();

  /** Destructor
  */
  ~Socket(){};

  /** close socket/disconnect from server
  * \return zero if socket was closed properly, -1 on error
  */
  int close();

  /**
  */
  bool valid() const;

  /** initialize from already existing socket
   *  \arg fd valid file-descriptor
   *  \warning initialization may fail, check isConnected() afterwards
   *  \warning using this constructor will result in loss of the socket origin!
   */
  void set_fd( int fd );

  /**
   */
  int get_fd() const { return _fd; }
	
  /**
  */
  void setAddress(const Address& addr) {_addr = addr; }

  /**
   */
  std::string getURL() const { return _addr.getURL(); }

  /**
   */
  const Address& getAddress() const {return _addr; }

  /**
   */
  Address& getAddress()  {return _addr; }

  /**
   * values possible for the argument opt of setSockOpt
   */
  enum {no_delay, clo_exec, reuseaddr};

  /**
   *  permit to manipulate the socket options TCP_NODELAY, FD_CLOEXEC, SO_REUSEADDR
   * \arg opt option (no_delay, clo_exec, reuseaddr)
   * \arg value
   */
  int setSockOpt(int opt,int value);

  /**
   *     print the values of the socket
   */
  std::ostream& print(std::ostream&) const;

  /**
   *     return the values of the socket
   */
  char * c_str();

protected:
  Address _addr;
  int _fd;
  bool _valid;
  friend class SocketServer;
};

inline std::ostream& operator << (std::ostream& o, const Socket& s )
{
  return s.print(o);
}



/*! \class Socket
 *  \brief a Socket is an endpoint for communication
 *
 * the class Socket encapulates the socket, adding functions for
 * connection, writing, reading, checking for data
 */
} //namespace SOCK

#endif

