/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 *
 *
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */



#ifndef __IOREDIR2_H
#define __IOREDIR2_H

#include <vector>
#include <inuk/acom>
#include <inuk/pth>
#include <inuk/sock>
#include <pthread.h>


class IOredir;

enum IOTypes
  {
    IOerr,
    IOout,
    IOin,
    IOtaktuk
  };


typedef struct 
{
  int32_t emet_id;
  char* emet_name;
  int8_t iotype;
  int32_t size;
  char* buffer;
}local_iobuf;


class output:public acom::ServiceAM
{
 public:
  output(IOredir* IO);
  void call(
	    acom::ComFailure::Code error_no,
		acom::Network::Receipt* receipt,
	    acom::Network::Header* header,
	    acom::IOVectEntry* vector,
	    acom::IOVectSize count,
      void* ctxt
	    ) throw(acom::ComFailure);
 private:
  IOredir* _IO;
};

class input:public acom::ServiceAM
{
 public:
  input(IOredir* IO);
  void call(
	    acom::ComFailure::Code error_no,
		acom::Network::Receipt* receipt,
	    acom::Network::Header* header,
	    acom::IOVectEntry* vector,
	    acom::IOVectSize count,
      void* ctxt
	    ) throw(acom::ComFailure);
 private:
  IOredir* _IO;
};


//class IOredir : public PTH::ThreadModel<IOredir>
class IOredir
{
 public:
  IOredir();
  virtual ~IOredir();

  int initialize(acom::Network* net);
  typedef enum{none,all,in,killed} flag;
  int commit();
  int commit(int cloexec,int verbose);
  void restore();
  flag status();
  void IO_on();
  void IO_off();
  void IO_kill();
  int env_export();
  int env_recovery();
  void terminate();
  void terminatecommit();

 
 protected:
  SOCK::SocketClient* konsole_in;
  SOCK::SocketClient* konsole_out;
  SOCK::SocketClient* konsole_err;

  SOCK::SocketClient* fd_in;
  SOCK::SocketClient* fd_out;
  SOCK::SocketClient* fd_err;
  int verbose;
  int _rank;
 private:
  //the IO routing thread
  PTH::Thread IO_thread;
  PTH::Mutex state_mutex;
  PTH::Condition cond_state;
  PTH::ThreadId IO_threadId;
  //	PTH::CountDown join;
  int once_control;
  void waiting(flag _f);
  void waitingwhile(flag _f);

  char* HostName;


  SOCK::SocketClient wakeupW;
  SOCK::SocketClient wakeupR;
  SOCK::SocketClient* _in;
  SOCK::SocketClient* _out;
  SOCK::SocketClient* _err;

  //!function which route messages
  //  friend void thread_IO(IOredir* IO);
  friend class output;
  friend class input;
  friend void thread_IO(IOredir* IO);
  void localIOprint(local_iobuf* iobuf);
  void localIOrecv(local_iobuf* iobuf);

  void run();
  
  flag state;
  acom::Network* _net;
  output* oservice;
  input* iservice;
  acom::Network::OutChannel* com_channel  ;
  int localIOprocess(IOTypes type, SOCK::SocketClient* local_s);
  int konsoleprocess();

};

#endif
