/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#include <iostream>
#include <fstream>
#include <string>
#include <stdarg.h>
#include <unistd.h>
#include <inuk/acom>
#include <inuk/pth>
#include <inuk/taktuk/taktukinstall.h>

class drawtree{
	public:
		drawtree(const char* treename){
			std::string filename = treename;
			filelock.init();
			filename = "/tmp/" + filename;
			filelock.lock();
			file.open(filename.c_str(),std::ofstream::trunc);
			if(file.is_open())
			{
				file<<"digraph "<<treename<<" {"<<std::endl;
				file<<"size=\"9,12\";"<<std::endl;
				file<<"ratio=auto;"<<std::endl;
				file<<"rotate=90;"<<std::endl;
			}
			else
				std::cerr<<"WARNING draw tree can't open file "<<treename<<std::endl;
			localnode=-1;
			filelock.unlock();
			localload=0.0;
		}
		void record(int node , int child){
			filelock.lock();
			if(file.is_open())
				file<<node<<" -> "<<child<<";"<<std::endl;
			filelock.unlock();
			//localnode=node;
		}
		void setlabel(int node, char* text)
		{
			filelock.lock();
			if(file.is_open())
				file<<node<<" [label=\""<<text<<"\\nrank:"<<node<<"\\nloadavg:"<<localload <<"\"];"<<std::endl;
			filelock.unlock();
		}
		void setload(int node, double load)
		{
			filelock.lock();
			//if(file.is_open()&&load>0.80)
			if(file.is_open())
			{
				file<<node<<" [style=filled];"<<std::endl;
				file<<node<<" [fillcolor="<<(load>0.67?"red":(load<0.33?"green":"yellow"))<<"];"<<std::endl;
				file<<node<<" [tooltip=\""<<load<<"\"];"<<std::endl;
				localload=load;
			}
			filelock.unlock();
		}
		void terminate()
		{
			filelock.lock();
			if(file.is_open())
			{
				file<<"}"<<std::endl;
				file.close();
			}
			filelock.unlock();
		}
		void print()
		{
			char cmd[512];
			sprintf(cmd,"%s/viewnet",TAKTUKBIN);
      system(cmd);	
		}
		~drawtree(){
			filelock.destroy();
		}
	private:
		std::ofstream file;
		int localnode;
		double localload;
		PTH::Mutex filelock;
};

class draw_net : public acom::ServiceAM {
	public:
		void call (AC::ComFailure::Code e,      // if any
				AC::Network::Receipt* receipt,
				AC::Network::Header* h,
				AC::IOVectEntry* iov,
				AC::IOVectSize   iov_sz,
				void* ctxt
				) throw (acom::ComFailure){
			char rhostname[124];
			double l=0.0;
			
			if (e == AC::ComFailure::OK){
				receipt->unpack(rhostname,124);
				receipt->unpack(&l,sizeof(double));
				if(_drawer!=NULL)
				{
					_drawer->setload(h->src,l);
					_drawer->setlabel(h->src,rhostname);
					for(int i=0;i<iov_sz;i++)
					{
						_drawer->record(h->src,*(acom::Network::LocalNodeNo*)iov[i].data);
						//_drawer->setlabel(rhostname);
					}
				}
			}
		}
		
		draw_net(acom::Network* net){
			_net_name=net->get_name();
			_net=net;
			_drawer=NULL;
			_root=NULL;
			acom::Service::Manager.bind(acom::ServiceManager::NETDRAW_LNS,1, (acom::Service*)this );
			if(_net->rank()==0)
				_drawer=new drawtree(_net_name.c_str());
			else
				_root = _net->get_default_route(0); // sent it to the root 
			if(gethostname(hostname,124)<0)
				std::cerr<<"WARNING draw tree can't read hostname"<<std::endl;
		}
		
		~draw_net(){
			if(_drawer!=NULL)
				delete(_drawer);
		}
		
		void record(){
			double load=0.0;
#ifndef PTH_USE_PTHREAD_IRIX
			getloadavg(&load,1);
#endif
			if(_net->rank()==0)
			{
				for(acom::Network::iter_known_nodes neighbour = _net->begin_known_nodes(); neighbour!=_net->end_known_nodes(); neighbour++)
				{
					_drawer->record(_net->rank(),neighbour->second->_node_rank);
				}
				_drawer->setload(_net->rank(),load);
				_drawer->setlabel(_net->rank(),hostname);
			}else
			{
				AC::ActiveMessage<draw_net> am (_net, this, 0);
				am->pack(hostname,124);
				am->pack(&load,sizeof(double));
				for(acom::Network::iter_known_nodes neighbour = _net->begin_known_nodes(); neighbour!=_net->end_known_nodes(); neighbour++)
				{
					am->put(&neighbour->second->_node_rank,sizeof(acom::Network::LocalNodeNo));
				}
				am.post(_root);
			}
		}

		void terminate()
		{
			_net->barrier();
			if(_drawer!=NULL)
				_drawer->terminate();
		}

		void print()
		{
			if(_drawer!=NULL)
				_drawer->print();
		}
	private:
		std::string _net_name;
		drawtree* _drawer;
		acom::Network* _net;
		acom::Network::OutChannel* _root;// channel to the root 
		char hostname[124];
};
