/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * 
 * 
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#ifndef __REDUCTION_H
#define __REDUCTION_H

#include <inuk/acom>
#include <pthread.h>
#include <inuk/taktuk/taktuknetwork.h>
#include <time.h>

#define TIMEOUT 6000

class Reduction:public acom::ServiceAM
{
 public:
  Reduction()
	{
		_net=0;
		_count=0;
		reductab=NULL;
		count_mutex.init();
		count_cond.init();
		reducetab_mutex.init();
		acom::Service::Manager.bind(acom::ServiceManager::TAKTUK_LNS,8, (acom::Service*)this);
  }
	
	~Reduction()
	{
        count_mutex.destroy();
        count_cond.destroy();
				reducetab_mutex.destroy();
  }
	
  void init(Taktuk::Network* net)
	{
		_net=net;
	}
	 
  void call(acom::ComFailure::Code error_no,
						acom::Network::Receipt* receipt,
						acom::Network::Header* header,
						acom::IOVectEntry* vector,
						acom::IOVectSize count,
						void* ctxt
						) throw(acom::ComFailure)
	{
		reducetab_mutex.lock();
		if(reductab==NULL)
		{
			reductab=(void**)malloc(_net->node_count()*sizeof(void*));
			for(int i=0;i<_net->node_count();i++)
				reductab[i]=NULL;
		}

		if(header->src==0)
		{
			for(int i=0;i<_net->node_count();i++)
				reductab[i]=vector[i].data;
		}else
		{
			if(count>0)
				reductab[header->src]=vector[0].data;
			else
				dbg.log("reduce","ERROR IN REDUCE RECEIVE ELT ON NODE 0\n");
		}
		reducetab_mutex.unlock();
		ack();
	}

	
	void** reduce(void* elt, int32_t eltsize, int* reduce_size)
	{

		int res=1;
		if(reduce_size!=NULL)
			*reduce_size=_net->node_count();
		
		reducetab_mutex.lock();
		if(reductab==NULL)
		{
			reductab=(void**)malloc(_net->node_count()*sizeof(void*));
			for(int i=0;i<_net->node_count();i++)
				reductab[i]=NULL;
		}

		if(_net->rank()==0)
			reductab[0]=memcpy(malloc(eltsize),elt,eltsize);
		
		reducetab_mutex.unlock();
		
		if(_net->node_count()==1&&_net->rank()==0)
		{
			return reductab;
		}
		dbg.log("reduce","!!!!!!!!!!!!!!!!!!!!!!!!!!!!!BEGIN REDUCE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
		if(_net->rank()==0)
		{
			waitfor(_net->node_count()-1);
			try
			{
				Taktuk::Network::OutChannel* com;
				com=_net->get_default_route(Taktuk::Network::ALL);
				acom::AllocWriteAndSignal<Reduction>aws (_net, this, 0);
				reducetab_mutex.lock();
				for(int i=0;i<_net->node_count();i++)
				{
					if(reductab[i]!=NULL)
						aws->put(reductab[i],eltsize);
				}
				reducetab_mutex.unlock();
				aws.post(com);
				dbg.log("reduce", "msg sent\n");
			}
			catch(...)
			{
				res= -1;
				std::cout<<"REDUCE ERROR"<<std::endl;
			}
		}
		else
		{
			 	
			try
			{
				Taktuk::Network::OutChannel* com;
				com=_net->get_default_route(0);
				acom::AllocWriteAndSignal<Reduction>aws (_net, this, 0);
				if(elt!=0)
					aws->put(elt,eltsize);
				aws.post(com);
				dbg.log("reduce", "msg sent\n");
			}
			catch(...)
			{
				res= -1;
				std::cout<<"REDUCE ERROR"<<std::endl;
			}
			waitfor(1);
			dbg.log("reduce", "Apres waitack \n");
		}
		
		if(res>0)
			dbg.log("reduce","!!!!!!!!!!!!!!!!!!!!!!!!END reduce [OK]!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
		else
			dbg.log("reduce","!!!!!!!!!!!!!!!!!!!!!!!!END reduce [FAILED]!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
		return reductab;
	}
	
 private:
  void waitfor(int c_nodes)
	{
		count_mutex.lock();
		_count = _count + c_nodes;
		
		while(_count > 0)
		{
			dbg.log("reduce","___________________Wait for %d nodes__________________________\n",_count);
			count_cond.wait(count_mutex);
		}
		_count=0;
		count_mutex.unlock();
	}

	void ack()
	{
		count_mutex.lock();
		_count--;
		dbg.log("reduce","________________________REDUCE COUNT -- = %d _________________\n",_count);
		count_cond.signalnothrow();
		count_mutex.unlock();
	}

	int _count;
  Taktuk::Network* _net;
	void** reductab;
  PTH::Mutex count_mutex;
  PTH::Mutex reducetab_mutex;
  PTH::Condition count_cond;
};


#endif /* __TAKTUKREDUCTION_H */
