/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * 
 * 
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#ifndef __TAKTUKNETWORK_H
#define __TAKTUKNETWORK_H




#include <inuk/taktuk/CoreTCP.h>
#include <inuk/taktuk/taktukrouter.h>
#include <inuk/acom>
#include <inuk/pth>
#include <inuk/sock>
#include <string>
#include <iostream>
#include <sys/time.h>
#include <inuk/taktuk/connector.h>
#include <inuk/taktuk/topology.h>
#include <inuk/taktuk/IOredir.h>


class Barrier;
class spawner;
class scheduler;
class steal;
class getinfo;


namespace Taktuk
{



class Network:public  acom::Network, public taktuk_router
{
 public:
  Network();
  ~Network();
  
  static const acom::Network::LocalNodeNo ALL= ~acom::Network::LocalNodeNo(0);
  /**
     initialize: initialize the taktuk network 
     - topology 
     - redirect the IO
     - save several services & connectors
     - start the router thread
     @param argc arguments count modified after taktuk network arguments will be removed
     @param argv arguments modified after taktuk arguments will be removed
     @return 0 if all the init sequence is OK, -1 if arguments parsing error, -2 if redirect IO error and -3 if routing thread create error.
  */
  int initialize(int* argc, char** argv[])throw (acom::NetworkError);

  int initialize(int* argc, char** argv[],acom::Network* net)throw (acom::NetworkError){return -1;}

/**
     initialize: initialize the taktuk network 
     - topology      
     - save several services & connectors
     - IO are redirected but will be displayed when the router thread will be started
     @param argc arguments count modified after taktuk network arguments will be removed
     @param argv arguments modified after taktuk network arguments will be removed
     @param _AMWS boolean to activate in the init Active messages and Write & signal capabilities
     @param _IO boolean to activate in init the IO redirection mecanism
     @param _cloexec boolean to set all internal socket used with CLOEXEC flag
     @param use new inuktitut argument specification ( -- taktuk -c fork ...)
     @return 0 if all the init sequence is OK, -1 if arguments parsing error, -2 if redirect IO error and -3 if routing thread create error.
  */
  int initialize(int* argc, char** argv[],bool _AMWS,bool _IO,bool _cloexec,bool new_arg_specif = false )throw (acom::NetworkError);

  /**
     commit the initialize after this call all is up AMWS capabilities on and IO on and all nodes specified spawned
  */
  int commit()throw (acom::NetworkError);

  typedef acom::Network::LocalNodeNo LocalNodeNo;
  typedef acom::Network::Request     Request;

  //std::string get_name( ) const{return "TAKTUK";}

  /**
     exec_taktuk: fork exec a taktuk program. Files descriptors are left connected and reuse by the new taktuk program this function never return on success
     @param argv standards arguments with argv[0]=program name,  argv[1...NULL] program's argument. Warning: argv must be terminated by NULL
     @return never return on success or return same error of execve Cf. man documentation 
  */
  void exec_taktuk(char *argv[]);

  /**
     terminate: terminate and wait for all module
     - flush all the fd
     - set the router to standby and wait for it
     - free mem , close fd , destroy mutex (destructor of topology, router, services_manager,...)
     - that's all folks
  */
  int terminate()throw (acom::NetworkError);


  /** Logical clock for timestamping request
   */
  //    PTH::AtomicCounter _comClock;  // ??

  protected :
    
   /** iovect entries in a cell
   */
  static const int _nb_entries=10; // iovect entries in a cell 
  /** packed block size
   */
  static const int _block_sz = 5000; // packed block size

  /** Logical clock for timestamping request
   */
  PTH::AtomicCounter _comClock;

  /** size of cached free blocks
   *     */
  static const int _block_threshold=50; // size of cached free blocks
  /** size of cached free blocks
  */
  static const int _iovect_threshold=50; // size of cached free blocks
  /** size of cached free blocks
  */
  static const int _request_threshold=50; // size of cached free blocks

  
  
  
  acom::ByteCellAllocator _bA;
  /** iovec allocator 
   */
  acom::IOVectCellAllocator _iovA;
  
  void initAllocators(){
      _comClock=0;
     _reqA.init(_request_threshold);
     _bA.init(_block_sz, _block_threshold  );
     _iovA.init(sizeof(acom::IOVectCell)+(_nb_entries*sizeof(acom::IOVectEntry)) ,
				_iovect_threshold    );
     };
 
 
//  class TaktukChannel;
//	class TaktukLocalChannel;

  class TaktukRequest : public acom::Network::Request, public PTH::Linkable<TaktukRequest>
    {
      protected :
			bool init(size_t bsz, acom::ByteCellAllocator* bufa,
		  int iovsz, acom::IOVectCellAllocator * iova,
		  Network * manager)
			{return acom::Network::Request::init(bsz,bufa,iovsz, iova,manager);}
      
      friend class TaktukChannel;
			friend class TaktukLocalChannel;
      friend class Taktuk::Network;
      friend class taktuk_router;
    };
 public:
  class TaktukReceipt : public acom::Network::Receipt
    {
      protected :
			void set_local_link( int local_link)
			{_local_link=acom::Network::LocalNodeNo(local_link); }
			bool init( acom::IOVectSeq * iovs , acom::IOVectEntry * iov, size_t sz, acom::Network* net, int local_link)
			{_local_link=acom::Network::LocalNodeNo(local_link); return acom::Network::Receipt::init(iovs,iov , sz,net);}
			acom::Network::LocalNodeNo _local_link;
//			acom::Network::LocalNodeNo get_reply_channelID(){
//				return _local_link;
//			}
      friend class TaktukChannel;
			friend class TaktukLocalChannel;
      friend class Taktuk::Network;
      friend class taktuk_router;
			friend class steal;
			friend class getinfo;
	public:	
			acom::Network::LocalNodeNo get_reply_channelID(){
				return _local_link;
			}
    };

  /** request allocator 
   */
  
  typedef PTH::SyncCellAllocator<TaktukRequest> RequestAllocator ;
  RequestAllocator _reqA;
  


 public:
  class TaktukChannel : public acom::Network::OutChannel
    {
    public:
      static const int critical=0;
      static const int normal=1;
      static const int free=2;

      TaktukChannel( Taktuk::Network* n, acom::Network::LocalNodeNo rank,InfoTCPnet* l)
	{
	  _net  = n;
	  _peer = rank;
	  _info = l;
	}

      void post( acom::Network::Request* req ) throw(acom::ComFailure);

    protected:
      InfoTCPnet* _info;
      Taktuk::Network* _taktuk;
    };
  friend class TaktukChannel;
	friend class steal;
	friend class getinfo;
	friend class TaktukReceipt;
	friend class TaktukLocalChannel;
 public:
	class TaktukLocalChannel : public TaktukChannel
		{
		public:
			TaktukLocalChannel( Taktuk::Network* n, acom::Network::LocalNodeNo rank,InfoTCPnet* l) : TaktukChannel(n,rank,l){}
			
      void post( acom::Network::Request* req ) throw(acom::ComFailure);
		};
 public:



  /**
     newarg: changes user's arguments for all nodes launched by this nodes
  */
  void newarg(int* argc, char** argv[]);

  /**
     newarg: changes user's arguments for all nodes launched by this nodes
  */
  void newarg(std::string new_arg);

  /**
     add_connector: add a connector to the collector WARNING call this function before a spawn or a commit call
     @param connect: a new connector defined by the user
     @param id: string id for this connector
  */
/*	template<class T>
  void add_connector(const std::string& id);
	{
		//Spawner->add<T>(id);
	}
*/

  /**
     spawn: spawn all the sub nodes describe in the topology, after this call the user code will be execute on all specified nodes
     @return -1 if one node could not be spawn, 0 if all is good
  */
  int spawn();
  /**
     print_rejected: print on stderr the list of host with connection rejected during the spawn
  */
  void print_rejected();

  /**
     barrier: synchronize all the node
  */
  void barrier()throw (acom::ComFailure);


  /**
     print a usage help to the stderr
  */
  void usage();

// ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Allocate a new request.
  */
   acom::Network::Request* allocate_request () throw (acom::BadAlloc) { 
			TaktukRequest* req = _reqA.alloc();
			if(!req) throw acom::BadAlloc("[[Taktuk::allocate_request ::]] fail")  ;
			if(!req->init( _block_sz,& _bA,_nb_entries ,& _iovA,(Network*)this ))
				{_reqA.free(req);throw acom::BadAlloc("[[Taktuk::allocate_request ::]] fail")  ;}
      return req;
	}

  /** free request descriptor when unused
  */
   void deallocate_request( acom::Network::Request* rq ) throw (acom::InvalidArgument) {
    TaktukRequest* req = (TaktukRequest*)rq;
		if(req){
			if(req->get_net()!=this) 
				throw acom::InvalidArgument("[[Taktuk::deallocate_request ::]] not a Taktuk request")  ;
			else
			  _reqA.free(req); return ;
			}
		else throw acom::InvalidArgument("[[Taktuk::deallocate_request ::]] null request")  ;
  };

  /** Cancel a request allocated by allocate_request ( true == cancel done)
      Cancelation of request depend on state of progress
      @param req (IN)
  */
  bool cancel_request( acom::Network::Request* req ) 
    throw (acom::InvalidArgument) 
  { return false; }


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Allocate  an immediate buffer of sz bytes in network memory
      @param  sz (IN) in bytes 
  */
   void* allocate_data( size_t sz ) 
       throw (acom::BadAlloc) { return ::malloc(sz);};

  /** Deallocate a previously allocated immediate buffer  in network memory
      @param  data (IN) byte block to deallocate
  */
   void deallocate_data(  void* data ) 
       throw (acom::InvalidArgument) { ::free(data);};

  // --------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  ACOM::Network::CallContext* allocate_callcontext(size_t s) throw (ACOM::BadAlloc);
  
  void deallocate_callcontext(ACOM::Network::CallContext* cc) throw (ACOM::InvalidArgument);

	ACOM::Network::CallContext* resize_callcontext(ACOM::Network::CallContext* cc, size_t sz ) throw (ACOM::BadAlloc);

	ACOM::Network::CallContext* acquire_callcontext() throw (ACOM::BadAlloc);

	void release_callcontext(ACOM::Network::CallContext* cc) throw (ACOM::InvalidArgument);

/*
  ACOM::Network::CallContext* resize_callcontext(ACOM::Network::CallContext* cc, size_t sz ) throw (ACOM::BadAlloc)
  {
    if (cc->_stack_size > sz) return cc;
    size_t offset = cc->size();
    ACOM::Network::CallContext*  new_cc = allocate_callcontext(sz+128);
    memcpy( new_cc->_base, cc->_base, offset );
    new_cc->_top = ((char*)new_cc->_base) + offset;
    deallocate_callcontext( cc );
    return new_cc;
  }
*/
  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Add to  network memory a part of user memory
      @param  begin (IN)IN) beginning address of the memory block 
      @param  sz sz   (IN)in bytes of the memory block 
      \warning The zone is mapped. The really mapped zone is the rounding of 
      the zone in system page size unit. At this level, no registering of 
      concurrent mappings or overlapping of zones are done.
  */
  void  map_zone( void * begin, size_t sz)throw (PTH::BadAlloc) {}

  /** Retract from network memory a part of user memory 
      @param  begin (IN) beginning address of the memory block 
      @param  sz   (IN) in bytes of the memory block 
      \warning The zone is unmapped. At this level, no control of overlapping 
      of mapping or multiples mapping of the same zone are done..   
  */
  void  umap_zone( void * begin, size_t sz)throw (acom::InvalidArgument) {}


  // ---------------------------------------------------------------------------  
  // ---------------------------------------------------------------------------  
  /** plug the local node in the network
       @param   nodename(IN) network identification of remote node
      \warning ajouter les aliasing d'adresse de noeud struct 
  */
  void bind_node( std::string nodename  )
    throw(acom::NetworkError){throw acom::NetworkError("Not Yet Implemented");}

     
  /** add a new node to the network
    @param  nodename(IN) network identification of remote node
    @return local node number of the new node
   \warning ajouter les aliasing d'adresse de noeud   
  */
  void add_node( std::string nodename , LocalNodeNo peer ) 
    throw(acom::NetworkError){throw acom::NetworkError("Not Yet implemented");}

   
  /** del a previously added node to the network
      @param  peer  local node Number of remote node to disconnect
      \warning ajouter les aliasing d'adresse de noeud     
  */
  void  del_node( LocalNodeNo  peer  ) 
       throw(acom::NetworkError){throw acom::NetworkError("Not Yet implemented");}


  // ---------------------------------------------------------------------------    
  // ---------------------------------------------------------------------------    
  /** Return the route (channel) to send data to LocalNodeNo peer
      @param   peer(IN) remote node id
  */
  TaktukChannel* get_default_route( LocalNodeNo peer )
       throw(acom::NetworkError);

  /** Set a channel as a default channel to a destination node
      @param   ch default channel 
  */
  void set_default_route(  acom::Network::OutChannel* ch  ) 
    throw(acom::NetworkError){throw acom::NetworkError("Not Yet implemented");}



  /** Open a channel between local node  to remote peer
      @param   peer(IN) remote node id
      \warning ajouter les aliasing d'adresse de noeud  
  */
  TaktukChannel* open_channel( LocalNodeNo peer ) 
       throw(acom::ComFailure){throw acom::NetworkError("Not Yet implemented");}

  /** Close a channel between local node  to remote peer
      @param   ch (IN)channel to close 
      \warning ajouter les aliasing d'adresse de noeud  
  */
  void close_channel( acom::Network::OutChannel* ch ) 
       throw(acom::ComFailure){throw acom::NetworkError("Not Yet implemented");}

  // ---------------------------------------------------------------------------  
  // ---------------------------------------------------------------------------  
  /** broadcast to nodes (AM,AWS,WS protocols) 
      @param   req  request (IN) 
  */
  void broadcast( Network::Request* req , Network::OutChannel* ch=0) 
    throw (acom::ComFailure){throw acom::ComFailure("Not Yet implemented");}
  
  /** neighbour_cast to nodes (AM,AWS,WS protocols) 
      @param   req  request (IN) 
  */
  void neighbour_cast( Network::Request* req, Network::OutChannel* ch=0 ) 
       throw (acom::ComFailure){throw acom::ComFailure("Not Yet implemented");}

  void  post( Network::Request* req, Network::OutChannel* ch=0 ) 
    throw (acom::ComFailure){throw acom::ComFailure("Not Yet implemented");}
  
 protected:

  //topolgy used to interconnected nodes
  topology topo;
  friend class taktuk_router;
	friend class scheduler;
 


 protected:
 
  //build the lauching tree
  spawner* Spawner;
 
  //perf time counter
  timeval time_init;
  timeval time_term; 
 public:
          void RefreschNetwork();
	PTH::Mutex _LocalChannel_mut;
	std::vector<TaktukLocalChannel*> LocalChannel; //copy of topo.info.all_links
 private:
	void InitNetwork();
//	void RefreschNetwork();
	int InitIO(SOCK::SocketClient** sc);
  std::string getsub(::topology::host* _host);
	//std::vector<TaktukLocalChannel*> LocalChannel; //copy of topo.info.all_links
	//PTH::Mutex _LocalChannel_mut;
  bool resuming;
  Barrier* _barrier;
	IOredir* _ioredir;
  bool _AMWS;
  bool _IOREDIR;
	PTH::Mutex _access;
};

}//end namespace Taktuk

#endif /*__TAKTUKNETWORK_H*/
